// Copyright 2017 Global Phasing Ltd.

// This file exists to make compilation of gemmi faster.

#define GEMMI_WRITE_IMPLEMENTATION
#include "gemmi/sprintf.hpp"
#include "gemmi/to_mmcif.hpp"
#include "gemmi/to_pdb.hpp"
#include "gemmi/mtz.hpp"
